"""
Log events to Elastic Search via bulk upload
"""

import datetime as _datetime
import json as _json
import logging as _logging
import socket as _socket


__version__ = '0.2'

LOG = _logging.getLogger(__name__)


DEFAULT = {
    'host': 'localhost',
    'port': 9700,
    'protocol': 'UDP',
    'index': 'log',
    'datestamp_index': True,
    'type': None,
    }


class Connection(object):
    """A socket connecting to Elastic Search

    Use a context manager for PEP 343's 'with' syntax:

    >>> with Connection(host='localhost', port=1234) as c:
    ...     c.send(message='hello!')
    """
    def __init__(self, host=None, port=None, protocol=None):
        if host is None:
            host = DEFAULT['host']
        if port is None:
            port = DEFAULT['port']
        if protocol is None:
            protocol = DEFAULT['protocol']
        self.host = host
        self.port = port
        if protocol == 'UDP':
            self.socket_type = _socket.SOCK_DGRAM
        else:
            raise NotImplementedError(protocol)
        self._sock = None

    def __enter__(self):
        self._sock = _socket.socket(_socket.AF_INET, self.socket_type)
        return self

    def __exit__(self, *exc_info):
        if self._sock is not None:
            try:
                self._sock.close()
            finally:
                self._sock = None

    def send(self, message):
        LOG.debug(message)
        self._sock.sendto(message, (self.host, self.port))


def log(index=None, type=None, sort_keys=False, **kwargs):
    """Log an arbitrary payload dictionary to Elastic Search

    Uses the default connection configuration.  If you need to
    override any of them, build your payload dict by hand and use
    emit() instead.

    You can optionally override the index and type of payload, for
    later filtering in Elastic Search.  This means that `index` and
    `type` are not available as payload keys.
    """
    kwargs['@timestamp'] = _datetime.datetime.utcnow().isoformat()
    kwargs['@version'] = 1
    return emit(payload=kwargs, index=index, type=type, sort_keys=sort_keys)


def emit(payload, index=None, datestamp_index=None, type=None,
         sort_keys=False, connection_class=Connection, **kwargs):
    """Send bulk-upload data to Elastic Search

    Uses the 'index' action to add or replace a document as necessary.

    http://www.elasticsearch.org/guide/reference/api/bulk/
    http://www.elasticsearch.org/guide/reference/api/bulk-udp/
    """
    if index is None:
        index = DEFAULT['index']
    if type is None:
        type = DEFAULT['type']
    if datestamp_index is None:
        datestamp_index = DEFAULT['datestamp_index']
    if datestamp_index:
        index = '-'.join([
            index,
            _datetime.date.today().strftime('%Y.%m.%d'),
            ])

    if type is None:
        LOG.error('You must set a type for {!r}'.format(payload))
        return

    index_data = {
        'index': {
            '_index': index,
            '_type': type,
            },
        }
    #not everything is JSON serializable, and logging should not blow up
    #how to make these errors easier to track down?
    try:
        message = '\n'.join([
            _json.dumps(index_data, sort_keys=sort_keys),
            _json.dumps(payload, sort_keys=sort_keys),
            '',
        ])
    except TypeError:
        LOG.error('Unable to serlialize {!r} to json'.format(payload))
        message = '\n'.join([
            _json.dumps(index_data, sort_keys=sort_keys),
            _json.dumps({"error": "unable to serialize"}, sort_keys=sort_keys),
            '',
        ])

    if hasattr(message, 'encode'):
        message = message.encode('utf-8')  # convert str to bytes for Python 3

    with connection_class(**kwargs) as connection:
        connection.send(message)

    return message



import unittest as _unittest
try:
    import unittest.mock as _mock
except ImportError:
    import mock as _mock

import bes as _bes
from . import udp_listener as _udp_listener


class ConnectionTestCase (_unittest.TestCase):
    def test_udp_connection(self):
        with _udp_listener.UDPListener(count=2) as listener:
            with _bes.Connection(
                    host=listener.host, port=listener.port, protocol='UDP'
                    ) as connection:
                connection.send(b'Hello!')
                connection.send(b'Goodbye!')
        self.assertEqual(
            [msg for msg,addr in listener.messages],
            [b'Hello!', b'Goodbye!'])


class EmitTestCase (_unittest.TestCase):
    def _call_emit(self, *args, **kwargs):
        return _bes.emit(
            datestamp_index=False,
            sort_keys=True,
            connection_class=_mock.MagicMock(),
            *args,
            **kwargs)

    def test_emit(self):
        message = self._call_emit(
            type='record',
            payload={'hello': 'world', 'goodbye': 'everybody'})
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "record"}}',
                b'{"goodbye": "everybody", "hello": "world"}',
                b'',
            ]))



import re as _re
import unittest as _unittest
try:
    import unittest.mock as _mock
except ImportError:
    import mock as _mock

try:
    from django.conf import settings as _settings
except ImportError as e:
    _settings_error = e
    _settings = None
else:
    if not _settings.configured:
        _settings.configure(
            BULK_ELASTIC_SEARCH_LOGGING_PORT=9999,
            BULK_ELASTIC_SEARCH_LOGGING_DATESTAMP_INDEX=False,
            # for contrib.auth
            DATABASES={
                'default': {
                    'ENGINE': 'django.db.backends.dummy',
                    },
                },
            )
    try:
        import django.http as _http
        import django.contrib.auth.models as _auth

        import bes.django as _bes_django
    except ImportError as e:
        _settings_error = None
        _settings = None

import bes as _bes


DATE_TIME_REGEXP = _re.compile(
    r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}[.]\d{6}')


def clean_message(message):
    if not hasattr(str, 'encode'):
        message = str(message, 'utf-8')

    search = DATE_TIME_REGEXP.search(message)
    if not search:
        raise ValueError('no timestamp in {!r}'.format(message))

    message = DATE_TIME_REGEXP.sub(
        'YYYY-MM-DDTHH:MM:SS.XXXXXX', message, count=1)
    return (message, search.group(0))


def skip_if_settings_is_none(obj):
    if _settings is None:
        return _unittest.skip(str(_settings_error))(obj)
    return obj


@skip_if_settings_is_none
class ConfigTestCase (_unittest.TestCase):
    def test_django_config_override_default(self):
        self.assertEqual(
            _bes.DEFAULT['port'],
            _settings.BULK_ELASTIC_SEARCH_LOGGING_PORT)


@skip_if_settings_is_none
class LogTestCase (_unittest.TestCase):
    def test_log_user_anonymous_request(self):
        request = _http.HttpRequest()
        request.user = _auth.AnonymousUser()
        message = _bes_django.log_user(request=request, sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "user-action"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "user_id": null, "username": ""}',
                b'',
            ]))

    def test_log_user_request(self):
        request = _http.HttpRequest()
        request.user = _auth.User(id=123, username='jdoe')
        message = _bes_django.log_user(request=request, sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "user-action"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "user_id": 123, "username": "jdoe"}',
                b'',
            ]))

    def test_log_user_request_override_type(self):
        request = _http.HttpRequest()
        request.user = _auth.User(id=123, username='jdoe')
        message = _bes_django.log_user(
            type='login', request=request, sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "login"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "user_id": 123, "username": "jdoe"}',
                b'',
            ]))

    def test_log_user_request_additional_data(self):
        request = _http.HttpRequest()
        request.user = _auth.User(id=123, username='jdoe')
        message = _bes_django.log_user(
            request=request, module='testing', sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "user-action"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "module": "testing", "user_id": 123, "username": "jdoe"}',
                b'',
            ]))

    def test_log_request_path(self):
        request = _http.HttpRequest()
        request.user = _auth.User(id=123, username='jdoe')
        request.path = '/music/bands/the_beatles/'
        request.META = {'QUERY_STRING': 'print=true'}
        message = _bes_django.log_request_path(request=request, sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "request"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "request_path": "/music/bands/the_beatles/?print=true", "user_id": 123, "username": "jdoe"}',
                b'',
            ]))

    def test_log_request_body(self):
        request = _http.HttpRequest()
        request.user = _auth.User(id=123, username='jdoe')
        request.path = '/music/bands/the_beatles/'
        request.META = {'QUERY_STRING': 'print=true'}
        request.read = lambda: b'Hey Jude'
        message = _bes_django.log_request_body(request=request, sort_keys=True)
        message, timestamp = clean_message(message)
        self.assertEqual(
            message,
            b'\n'.join([
                b'{"index": {"_index": "log", "_type": "request"}}',
                b'{"@timestamp": "YYYY-MM-DDTHH:MM:SS.XXXXXX", "@version": 1, "request_body": "Hey Jude", "request_path": "/music/bands/the_beatles/?print=true", "user_id": 123, "username": "jdoe"}',
                b'',
            ]))



from __future__ import absolute_import

from django.conf import settings as _django_settings

import bes as _bes


if _django_settings.configured:
    for key,value in _bes.DEFAULT.items():
        django_key = 'BULK_ELASTIC_SEARCH_LOGGING_{}'.format(key.upper())
        _bes.DEFAULT[key] = getattr(_django_settings, django_key, value)


def log_user(type='user-action', request=None, **kwargs):
    """Log activity requested by a Django user

    This uses request.user.id and request.user.username.  That's what
    you get with the default contrib.auth.User, but likely you'll have
    them even if you override AUTH_USER_MODEL.

    https://docs.djangoproject.com/en/dev/ref/request-response/
    https://docs.djangoproject.com/en/dev/topics/auth/default/#user-objects
    """
    return _bes.log(
        type=type,
        user_id=request.user.id,
        username=request.user.username,
        **kwargs)


def log_request_path(type='request', request=None, **kwargs):
    """Like log_user, but also adds the request path
    """
    return log_user(
        type=type,
        request=request,
        request_path=request.get_full_path(),
        **kwargs)


def log_request_body(request=None, **kwargs):
    """Like log_request_path, but also adds the request body
    """
    return log_request_path(
        request=request,
        request_body=request.read(),
        **kwargs)
